@@ -2,11 +2,11 @@ import hashlib
 import logging
 import os
 import tempfile
-from typing import Dict, Optional
+from typing import Dict, Optional, TextIO, Union
 
 import esptool
 from pytest_embedded.log import PexpectProcess
-from pytest_embedded_serial_esp.serial import EspSerial
+from pytest_embedded_serial_esp.serial import EspSerial, EsptoolVersion
 
 from .app import IdfApp
 
@@ -31,11 +31,13 @@ class IdfSerial(EspSerial):
         skip_autoflash: bool = False,
         port_app_cache: Dict[str, str] = None,
         confirm_target_elf_sha256: bool = False,
+        erase_nvs: bool = False,
         **kwargs,
     ) -> None:
         self._port_app_cache: Dict[str, str] = port_app_cache if port_app_cache is not None else {}
         self.app = app
         self.confirm_target_elf_sha256 = confirm_target_elf_sha256
+        self.erase_nvs = erase_nvs
 
         if not hasattr(self.app, 'target'):
             raise ValueError(f'Idf app not parsable. Please check if it\'s valid: {self.app.binary_path}')
@@ -76,17 +78,10 @@ class IdfSerial(EspSerial):
             self.flash()
 
     @EspSerial.use_esptool
-    def flash(self, erase_nvs: bool = True) -> None:
+    def flash(self) -> None:
         """
         Flash the `app.flash_files` to the dut
-
-        Args:
-            erase_nvs: erase non-volatile storage blocks
         """
-        if not self.app.partition_table:
-            logging.error('Partition table not parsed. Skipping auto flash...')
-            return
-
         if not self.app.flash_files:
             logging.error('No flash files detected. Skipping auto flash...')
             return
@@ -95,10 +90,22 @@ class IdfSerial(EspSerial):
             logging.error('No flash settings detected. Skipping auto flash...')
             return
 
-        flash_files = [
-            (offset, open(path, 'rb')) for (offset, path, encrypted) in self.app.flash_files if not encrypted
-        ]
-        encrypt_files = [(offset, open(path, 'rb')) for (offset, path, encrypted) in self.app.flash_files if encrypted]
+        flash_files = [(file.offset, open(file.file_path, 'rb')) for file in self.app.flash_files if not file.encrypted]
+        encrypt_files = [(file.offset, open(file.file_path, 'rb')) for file in self.app.flash_files if file.encrypted]
+
+        nvs_file = None
+        if self.erase_nvs:
+            address = self.app.partition_table['nvs']['offset']
+            size = self.app.partition_table['nvs']['size']
+            nvs_file = tempfile.NamedTemporaryFile(delete=False)
+            nvs_file.write(b'\xff' * size)
+            if not isinstance(address, int):
+                address = int(address, 0)
+
+            if self.app.flash_settings['encrypt']:
+                encrypt_files.append((address, open(nvs_file.name, 'rb')))
+            else:
+                flash_files.append((address, open(nvs_file.name, 'rb')))
 
         # fake flasher args object, this is a hack until
         # esptool Python API is improved
@@ -118,34 +125,23 @@ class IdfSerial(EspSerial):
             'ignore_flash_encryption_efuse_setting': False,
             'erase_all': False,
         }
+
+        if self.ESPTOOL_VERSION == EsptoolVersion.V4:
+            default_kwargs['force'] = False
+
         default_kwargs.update(self.app.flash_settings)
-        flash_args = FlashArgs(default_kwargs)
-
-        nvs_file = None
-        if erase_nvs:
-            address = self.app.partition_table['nvs']['offset']
-            size = self.app.partition_table['nvs']['size']
-            nvs_file = tempfile.NamedTemporaryFile(delete=False)
-            nvs_file.write(b'\xff' * size)
-            if not isinstance(address, int):
-                address = int(address, 0)
-
-            if self.app.flash_settings['encrypt']:
-                encrypt_files.append((address, open(nvs_file.name, 'rb')))
-            else:
-                flash_files.append((address, open(nvs_file.name, 'rb')))
+        default_kwargs.update(self.app.flash_args.get('extra_esptool_args', {}))
+        args = FlashArgs(default_kwargs)
 
         try:
             if self.proc.baudrate < self.SUGGEST_FLASH_BAUDRATE:
                 self.stub.change_baud(self.SUGGEST_FLASH_BAUDRATE)
 
-            esptool.detect_flash_size(self.stub, flash_args)
-            esptool.write_flash(self.stub, flash_args)
+            esptool.detect_flash_size(self.stub, args)
+            esptool.write_flash(self.stub, args)
 
             if self.proc.baudrate > self.DEFAULT_BAUDRATE:
                 self.stub.change_baud(self.DEFAULT_BAUDRATE)  # set to the default one to get the serial output
-        except Exception:  # noqa
-            raise
         finally:
             if nvs_file:
                 nvs_file.close()
@@ -161,21 +157,24 @@ class IdfSerial(EspSerial):
     @EspSerial.use_esptool
     def dump_flash(
         self,
-        output_filepath: str,
         partition: Optional[str] = None,
         address: Optional[str] = None,
         size: Optional[str] = None,
-    ) -> None:
+        output: Union[str, TextIO, None] = None,
+    ) -> Optional[bytes]:
         """
         Dump the flash bytes into the output file by partition name or by start address and size.
 
         Args:
-            output_filepath: output file path
+            output: file path or file stream to write to. File stream should be opened with bytes mode.
             partition: partition name
             address: address that start reading from
             size: read size
+
+        Returns:
+            None if `output` is `str` or file stream.
+            bytes if `output` is None.
         """
-        os.makedirs(os.path.dirname(output_filepath), exist_ok=True)
         if partition:
             partition = self.app.partition_table[partition]
             _addr = partition['offset']
@@ -187,8 +186,42 @@ class IdfSerial(EspSerial):
             raise ValueError('You must specify "partition" or ("address" and "size") to dump flash')
 
         content = self.stub.read_flash(_addr, _size)
-        with open(output_filepath, 'wb') as f:
-            f.write(content)
+        if output:
+            if isinstance(output, str):
+                os.makedirs(os.path.dirname(output), exist_ok=True)
+                with open(output, 'wb') as f:
+                    f.write(content)
+            else:
+                output.write(content)
+        else:
+            return content
+
+    @EspSerial.use_esptool
+    def erase_partition(self, partition_name: str) -> None:
+        """
+        Erase the partition provided
+
+        Args:
+            partition_name: partition name
+        """
+        if not self.app.partition_table:
+            raise ValueError('Partition table not parsed.')
+
+        if partition_name in self.app.partition_table:
+            address = self.app.partition_table[partition_name]['offset']
+            size = self.app.partition_table[partition_name]['size']
+            logging.info(f'Erasing the partition "{partition_name}" of size {size} at {address}')
+            self.stub.erase_region(address, size)
+        else:
+            raise ValueError(f'partition name "{partition_name}" not found in app partition table')
+
+    @EspSerial.use_esptool
+    def erase_flash(self) -> None:
+        """
+        Erase the complete flash
+        """
+        logging.info('Erasing the flash')
+        self.stub.erase_flash()
 
     @EspSerial.use_esptool
     def read_flash_elf_sha256(self) -> bytes:
@@ -216,6 +249,10 @@ class IdfSerial(EspSerial):
         Returns:
             True if the sha256 values are matched
         """
+        if not self.app.elf_file:
+            logging.info('no elf file. Can\'t tell if the target flashed the same elf file or not. Assume as False')
+            return False
+
         flash_elf_sha256 = self.read_flash_elf_sha256()
         elf_sha256 = hashlib.sha256()
         with open(self.app.elf_file, 'rb') as fr:
